	TITLE	'IDE MODULE FOR CP/M 3.0'

	; IDE3.ASM  v.1.0a  03/03/2023

	; Written by Wayne Parham  (wayne@parhamdata.com)

	; IDE disk device driver module for Altair computers.
	; Supports the IDE-CF controller provided by
	; S100computers.com.

	; John Monahan designed the controller board and wrote
	; a version of BIOS for his Z80-based systems.  The IDE
	; interface was based on a design from Peter Faasse.
	; Another important contributor is David Fry who recognized
	; that the original format left "holes" on the disk because
	; it skipped sectors in the LBA translation.

	; This version is a re-write that takes all of those things
	; and incorporates them in an 8080-compatible format that
	; runs on an original Altair and compatible computers.

TRUE		EQU	-1
FALSE		EQU	NOT TRUE

BANKED		EQU	FALSE	;<--------------- NON-BANKED VERSION

;INCLUDE CP/M 3.0 DISK DEFINITION MACROS:
MACLIB	CPM3

;INCLUDE Z-80 MACRO LIBRARY:
MACLIB	Z80

;Ports for 8255 chip. Change these to specify where the 8255 is addressed,
;and which of the 8255's ports are connected to which IDE signals.
;The first three control which 8255 ports have the control signals,
;upper and lower data bytes.  The last one is for mode setting for the
;8255 to configure its ports, which must correspond to the way that
;the first three lines define which ports are connected.

IDEportA	EQU	030H		;lower 8 bits of IDE interface
IDEportB	EQU	031H		;upper 8 bits of IDE interface
IDEportC	EQU	032H		;control lines for IDE interface
IDEportCtrl	EQU	033H		;8255 configuration port
IDEDrive	EQU	034H		;drive select

READcfg8255	EQU	10010010b	;set 8255 IDEportC out, IDEportA/B input
WRITEcfg8255	EQU	10000000b	;set all three 8255 ports output

;IDE control lines for use with IDEportC.  Change these 8
;constants to reflect where each signal of the 8255 each of the
;IDE control signals is connected.  All the control signals must
;be on the same port, but these 8 lines let you connect them to
;whichever pins on that port.

IDEa0line	EQU	01H	;direct from 8255 to IDE interface
IDEa1line	EQU	02H	;direct from 8255 to IDE interface
IDEa2line	EQU	04H	;direct from 8255 to IDE interface
IDEcs0line	EQU	08H	;inverter between 8255 and IDE interface
IDEcs1line	EQU	10H	;inverter between 8255 and IDE interface
IDEwrline	EQU	20H	;inverter between 8255 and IDE interface
IDErdline	EQU	40H	;inverter between 8255 and IDE interface
IDErstline	EQU	80H	;inverter between 8255 and IDE interface
;
;Symbolic constants for the IDE Drive registers, which makes the
;code more readable than always specifying the address pins

REGdata		EQU	IDEcs0line
REGerr		EQU	IDEcs0line + IDEa0line
REGseccnt	EQU	IDEcs0line + IDEa1line
REGsector	EQU	IDEcs0line + IDEa1line + IDEa0line
REGcylinderLSB	EQU	IDEcs0line + IDEa2line
REGcylinderMSB	EQU	IDEcs0line + IDEa2line + IDEa0line
REGshd		EQU	IDEcs0line + IDEa2line + IDEa1line		;(0EH)
REGcommand	EQU	IDEcs0line + IDEa2line + IDEa1line + IDEa0line	;(0FH)
REGstatus	EQU	IDEcs0line + IDEa2line + IDEa1line + IDEa0line
REGcontrol	EQU	IDEcs1line + IDEa2line + IDEa1line
REGastatus	EQU	IDEcs1line + IDEa2line + IDEa1line + IDEa0line

;IDE Command Constants.  These should never change.

COMMANDrecal	EQU	10H
COMMANDread	EQU	20H
COMMANDwrite	EQU	30H
COMMANDinit	EQU	91H
COMMANDid	EQU	0ECH
COMMANDspindown	EQU	0E0H
COMMANDspinup	EQU	0E1H
;
MAXSEC		EQU	3FH
LBA$MODE	EQU	11100000b
;
; IDE Status Register:
;  bit 7: Busy	1=busy, 0=not busy
;  bit 6: Ready 1=ready for command, 0=not ready yet
;  bit 5: DF	1=fault occurred
;  bit 4: DSC	1=seek complete
;  bit 3: DRQ	1=data request ready, 0=not ready to xfer yet
;  bit 2: CORR	1=correctable error occurred
;  bit 1: IDX	vendor specific
;  bit 0: ERR	1=error occurred
;
BUSY$FLAG	EQU	10000000b
BUSY$DRQ	EQU	10001000b
RDY$DRQ		EQU	00001000b
BUSY$CMD	EQU	11000000b	
RDY$CMD		EQU	01000000b

; Seconds to wait for drive to become ready after power-up
INIT$DELAY	EQU	2


;------------------------------------------------------------------
; Console ASCII equates
;------------------------------------------------------------------
BELL		EQU	07H
CR		EQU	0DH
LF		EQU	0AH


;------------------------------------------------------------------
; Disk assignment equates
;------------------------------------------------------------------
;
; Hardware-defined "relative" addressing values for IDE controller
;
RDRV0		EQU	0	;First drive on controller
RDRV1		EQU	1	;Second drive on controller	


;------------------------------------------------------------------
; Disk Parameter Headers
;------------------------------------------------------------------
	; Public labels:
	PUBLIC	DPH2, DPH3	;IDE drive parameter headers

   IF BANKED
	DSEG			;Disk Parameter Headers can be banked
   ENDIF

	; External labels:
	EXTRN	@ADRV,@RDRV
	EXTRN	@DMA,@TRK,@SECT
	EXTRN	@CBNK
	EXTRN	@DBNK
	EXTRN	@ERMDE		;BDOS error mode
	EXTRN	?WBOOT
	EXTRN	?PMSG		;Print null-terminated string
	EXTRN	?PDERR		;Print BIOS disk error
	EXTRN	?CONIN,?CONO	;Console I/O
	EXTRN	?CONST		;Console status
	EXTRN	?BNKSL		;Select memory bank

	; Extended disk parameter header for IDE drive 0:
	DW	HDWRT		;IDE Drive Write
	DW	HDRD		;IDE Drive Read
	DW	LOGIN0		;IDE Login
	DW	INIT0		;IDE Init
	DB	RDRV0		;Controller drive number
	DB	0		;Media Type

DPH2:	DPH	0,IDEHD$DPB,0,

	; Extended disk parameter header for IDE drive 1:
	DW	HDWRT		;IDE Drive Write
	DW	HDRD		;IDE Drive Read
	DW	LOGIN1		;IDE Login
	DW	INIT1		;IDE Init
	DB	RDRV1		;Controller drive number
	DB	0		;Media Type

DPH3:	DPH	0,IDEHD$DPB,0,

	CSEG			;Disk Parameter Blocks need to be in common memory

IDEHD$DPB:
	DPB	512,64,256,2048,1024,1,8000H

	; 512 x 64 x 256 = 8,388,608 bytes
	;
	; 512 byte sectors, 64 sectors per track, 256 tracks per disk
	; 2048 byte allocation unit, 1024 directory entries
	; 1 reserved system track, non-removable media


   IF BANKED
	DSEG			;Driver interface code can reside in banked memory
   ENDIF

;------------------------------------------------------------------
; Drive Login (not needed)
;------------------------------------------------------------------

LOGIN0:
LOGIN1:
	RET


;------------------------------------------------------------------
; IDE Initialization
;------------------------------------------------------------------

INIT0:
	MVI	A, RDRV0	;Select drive 0
	OUT	IDEDrive
	JMP	IDEinit

INIT1:
	MVI	A, RDRV1	;Select drive 1
	OUT	IDEDrive

IDEinit:			;Initialize the 8255 and drive
	MVI	A, READcfg8255
	OUT	IDEportCtrl	;config 8255 chip, READ mode
				
	MVI	A, IDErstline
	OUT	IDEportC	;reset

	MVI	B, 20H
ResetDelay:
	DCR	B
	JNZ	ResetDelay	;delay (reset pulse width)

	XRA	A
	OUT	IDEportC	;no IDE control lines asserted (just bit 7 of port C)
	CALL	DELAY$32
			
	MVI	D, LBA$MODE	;11100000B

	MVI	E, REGshd  
	CALL	IDEwr8D		;write byte to select the MASTER device

	MVI	B, INIT$DELAY	;set delay time (seconds)
WaitInit:	
	MVI	E, REGstatus	;get status after initialization
	CALL	IDErd8D		;check status
	MOV	A, D
	ANI	BUSY$FLAG	;check busy flag
	JZ	DoneInit	;return if ready bit is zero

	PUSH	B
	LXI	B, 0FFFFH	;long delay to allow drive to get up to speed	
DELAY2:	MVI	D, 2
DELAY1:	DCR	D
	JNZ	DELAY1
	DCX	B
	MOV	A, C
	ORA	B
	JNZ	DELAY2
	POP	B
	DCR	B
	JNZ	WaitInit

	CALL	SetErrorFlag	;if not ready by this point,

	LXI	H, msgDrive	;report "Drive unavailable"
	CALL	?PMSG
	LDA	@ADRV
	ADI	'A'		;convert drive number to letter
	MOV	C, A
	CALL	?CONO
	LXI	H, msgUnavailable
	CALL	?PMSG

	ORI	1		;return with error
	RET	

DoneInit:			;if drive reorts ready
	XRA	A		;return with no error
	RET


;------------------------------------------------------------------
; SECTOR WRITE ROUTINE
;------------------------------------------------------------------

HDWRT:				;Write one sector
	LDA	@RDRV
	OUT	IDEDrive	;select current drive
	XRA	A
	STA	ERFLG		;clear error flag
	CALL	wrlba		;convert track/sector to LBA
	CALL	IDEwaitnotbusy	;make sure drive is ready
	JC	SetErrorFlag

	MVI	D, COMMANDwrite
	MVI	E, REGcommand
	CALL	IDEwr8D		;tell drive to write a sector
	CALL	IDEwaitdrq	;wait unit it wants the data
	JC	SetErrorFlag	;if problem abort

	LHLD    @DMA		;DMA address

   IF BANKED			;write from banked memory block
	JMP	SW$WR$BNK
	CSEG			;put code block below in common memory
SW$WR$BNK:
	LDA	@CBNK
	PUSH	PSW		;save common bank address
	LDA	@DBNK
	CALL	?BNKSL		;switch to banked memory
   ENDIF

	MVI	A, WRITEcfg8255
	OUT	IDEportCtrl

	MVI	B,0		;256x2 bytes
WRSEC1:
	MOV	A, M
	INX	H
	OUT	IDEportA	;LOW byte to port A first
	MOV	A, M
	INX	H
	OUT	IDEportB	;then HIGH byte to port B
	MVI	A, REGdata
	PUSH	PSW
	OUT	IDEportC	;send write command
	ORI	IDEwrline	;send WR pulse
	OUT	IDEportC
	POP	PSW
	OUT	IDEportC
	DCR	B
	JNZ	WRSEC1
	MVI	A, READcfg8255	;set 8255 back to read mode
	OUT	IDEportCtrl

   IF BANKED			;switch back to common bank 
	POP	PSW
	CALL	?BNKSL
	JMP	CHECK$RW
	DSEG			;following code can be in banked memory
   ENDIF

CHECK$RW:
	MVI	E, REGstatus	;check R/W status when done
	CALL	IDErd8D
	MOV	A, D
	ANI	01H
	STA	ERFLG		;return Z if successful
	RZ			;but if not zero,

SetErrorFlag:			;return with error flag set
	XRA	A
	DCR	A
	STA	ERFLG
	RET


;------------------------------------------------------------------
; SECTOR READ ROUTINE
;------------------------------------------------------------------

HDRD:				;Read one sector
	LDA	@RDRV
	OUT	IDEDrive	;select current drive
	XRA	A
	STA	ERFLG	     	;clear error flag
	CALL	wrlba		;convert track/sector to LBA
	CALL	IDEwaitnotbusy	;make sure drive is ready
	JC	SetErrorFlag	;NZ set if error

	MVI	D, COMMANDread
	MVI	E, REGcommand
	CALL	IDEwr8D		;send sector read command
	CALL	IDEwaitdrq	;wait until data ready
	JC	SetErrorFlag	;if problem abort
		
	LHLD  	@DMA		;DMA address

   IF BANKED			;read into banked memory block
	JMP	SW$RD$BNK
	CSEG			;put code block below in common memory
SW$RD$BNK:
	LDA	@CBNK
	PUSH	PSW		;save common bank address
	LDA	@DBNK
	CALL	?BNKSL		;switch to banked memory
   ENDIF

	MVI	B, 0
MoreRD16:
	MVI	A,REGdata	;select data register address
	OUT	IDEportC	

	ORI	IDErdline	;08H+40H, pulse RD line
	OUT	IDEportC	

	IN	IDEportA	;read the LOW byte first
	MOV	M, A
	INX	H
	IN	IDEportB	;then the HIGH byte
	MOV	M,A
	INX	H
	
	MVI	A, REGdata	;deassert RD line
	OUT	IDEportC

	DCR	B
	JNZ	MoreRD16

   IF BANKED			;switch back to common bank 
	POP	PSW
	CALL	?BNKSL
   ENDIF	

	JMP	CHECK$RW

   IF BANKED
	DSEG			;following code can be in banked memory
   ENDIF


;------------------------------------------------------------------
; SUPPORT ROUTINES
;------------------------------------------------------------------

wrlba:				;Translate to logical block address
	CALL	IDEwaitnotbusy	;make sure drive isn't busy
	LHLD	@TRK		;get the CP/M requested track high and low
	MOV	A, L		;get low byte of track, interested in bottom two bits
	RRC			;shift 'em
	RRC			;shift again to the top
	ANI	0C0H		;examine those two bits
	MOV	C, A		;save in [C]
	LDA	@SECT		;sector number (low byte) into [A]
	ANI	03FH		;take only bottom 6 bits
	ORA	C		;add in top 2 bits of track
	MOV	D, A		;send info to the drive
	MVI	E, REGsector
	CALL	IDEwr8D

	MOV	A, L		;get low byte of track again
	RRC
	RRC
	ANI	03FH
	MOV	C, A		;save in [C]
	MOV	A, H		;get high byte of track
	RRC			;rotate twice, leaving low 2 bits
	RRC			;in upper bits of [A]
	ANI	0C0H		;mask all but the two bits we want
	ORA	C		;add in the top 6 bits of the first track byte
	MOV	D, A		;send Low TRK#
	MVI	E, REGcylinderLSB
	CALL	IDEwr8D
	
	MOV	A, H		;get high byte of track
	RRC			;just the top 6 bits
	RRC
	ANI	03FH
	MOV	D, A		;send High TRK#
	MVI	E, REGcylinderMSB
	CALL	IDEwr8D

	MVI	D, 1		;one sector at a time
	MVI	E, REGseccnt
	CALL	IDEwr8D
	RET

IDEwaitnotbusy:			;Drive READY if 01000000
	MVI	B, 0FFH
	MVI	C, 0FFH		;delay, must be above 80H. Leave longer for slower drives
MoreWait:
	MVI	E, REGstatus	;wait for RDY bit to be set
	CALL	IDErd8D
	MOV	A, D
	ANI	BUSY$CMD	;11000000B
	XRI	RDY$CMD		;01000000B
	JZ	DoneNotbusy
	DCR	B
	JNZ	MoreWait
	DCR	C
	JNZ	MoreWait
	STC			;set carry to indicqate an error
	ret
DoneNotBusy:
	ORA	A		;clear carry it indicate no error
	RET

IDEwaitdrq:
	MVI	B, 0FFH
	MVI	C, 0FFH		;delay must be above 80H. Leave longer for slower drives
MoreDRQ:
	MVI	E, REGstatus	;dait for DRQ bit to be set
	CALL	IDErd8D
	MOV	A, D
	ANI	BUSY$DRQ	;10001000B
	CPI	RDY$DRQ		;00001000B
	JZ	DoneDRQ
	DCR	B
	JNZ	MoreDRQ
	DCR	C
	JNZ	MoreDRQ
	STC			;set carry to indicate error
	RET
DoneDRQ:
	ORA	A		;clear carry
	RET

DELAY$32:
	MVI	A, 40		;delay ~32ms
DELAY3:
	MVI	B, 0
M0:
	DCR	B
	JNZ	M0
	DCR	A
	JNZ	DELAY3 
	RET


;------------------------------------------------------------------
; Low Level 8 bit R/W to the drive controller.  These are the routines that talk
; directly to the drive controller registers, via the 8255 chip.  
; Note the 16 bit I/O to the drive (which is only for SEC R/W) is done directly 
; in the routines HDRD & HDWRT for speed reasons.
;------------------------------------------------------------------

IDErd8D:			;Read byte from IDE register in [E], return info in [D]
	MOV	A, E
	OUT	IDEportC	;drive address onto control lines

	ORI	IDErdline	;RD pulse pin (40H)
	OUT	IDEportC	;assert read pin

	IN	IDEportA
	MOV	D, A		;return with data in [D]

	MOV	A, E
	OUT	IDEportC	;deassert RD pin first

	XRA	A
	OUT	IDEportC	;zero all port C lines
	RET

IDEwr8D:			;Write byte in [D] to IDE register in [E]
	MVI	A, WRITEcfg8255	;set 8255 to write mode
	OUT	IDEportCtrl

	MOV	A, D		;get data put it in 8255 A port
	OUT	IDEportA

	MOV	A, E		;select IDE register
	OUT	IDEportC

	ORI	IDEwrline	;lower WR line
	OUT	IDEportC

	MOV	A, E
	OUT	IDEportC	;deassert WR pin first

	XRA	A		;deselect all lines including WR line
	OUT	IDEportC

	MVI	A, READcfg8255	;config 8255 chip, read mode on return
	OUT	IDEportCtrl
	RET


;----------------------------------------------------------------------------
; Module data
;----------------------------------------------------------------------------

msgDrive	DB	'Drive ',0
msgUnavailable	DB	' unavailable.',CR,LF,0
ERFLG		DB	0H

	END
