/* -------------------------------------------------------------------------- */
/* IDE Test Program                                                           */
/*                                                                            */
/* Wayne Parham                                                               */
/*                                                                            */
/* wayne@parhamdata.com                                                       */
/* -------------------------------------------------------------------------- */


#include "stdio.h"

/*
;-------------------------------------------------------------------------------
;IDE Interface equates:
;-------------------------------------------------------------------------------
;Ports for 8255 chip. Change these to specify where your 8255 is addressed,
;The first three control which 8255 ports have the control signals,
;upper and lower data bytes.  The last one (IDEportCtrl), is for mode setting
;for the 8255 to configure its actual I/O ports (A,B & C).  
;
;Note most drives these days don't use the old Head,Track, Sector terminology.
;Instead we use "Logical Block Addressing" or LBA. This is what we use below.
;LBA treats the drive as one continous set of sectors, 0,1,2,3,... 3124,...etc.
;However as seen below we need to convert this LBA to heads,tracks and sectors
;to be compatible with CP/M & MSDOS.
;
;-------------------------------------------------------------------------------
*/

#define FALSE            0x00
#define TRUE             0xFF

#define quiet            0x00
#define verbose          0xFF

#define IDEpA            0x30              /* Lower 8 bits of IDE interface   */
#define IDEpB            0x31              /* Upper 8 bits of IDE interface   */
#define IDEpC            0x32              /* Control lines for IDE interface */
#define IDEpCtrl         0x33              /* 8255 configuration port         */
#define IDEDrSel         0x34              /* Bit zero chooses drive 0 or 1   */

#define READcfg8255      0x92              /* IDE ports A & B input, C output */
#define WRITEcfg8255     0x80              /* IDE ports A, B and C output     */

#define IDEa0Line        0x01              /* IDE bus A0                      */
#define IDEa1Line        0x02              /* IDE bus A1                      */
#define IDEa2Line        0x04              /* IDE bus A2                      */
#define IDEcs0line       0x08              /* IDE bus /CS0                    */
#define IDEcs1Line       0x10              /* IDE bus /CS1                    */
#define IDEwrLine        0x20              /* IDE bus /WR                     */
#define IDErdLine        0x40              /* IDE bus /RD                     */
#define IDErstLine       0x80              /* IDE bus /RESET                  */

#define REGdata          0x08              /* /CS0                            */
#define REGerr           0x09              /* /CS0 + A0                       */
#define REGsecCnt        0x0A              /* /CS0 + A1                       */
#define REGsector        0x0B              /* /CS0 + A1 + A0                  */
#define REGcylLSB        0x0C              /* /CS0 + A2                       */
#define REGcylMSB        0x0D              /* /CS0 + A2 + A0                  */
#define REGshd           0x0E              /* /CS0 + A2 + A1                  */
#define REGcmd           0x0F              /* /CS0 + A2 + A1 + A0             */
#define REGstat          0x0F              /* /CS0 + A2 + A1 + A0             */
#define REGctrl          0x16              /* /CS1 + A2 + A1                  */
#define REGaStat         0x17              /* /CS1                            */

#define CMDrecal         0x10              /* Recalibrate command             */
#define CMDread          0x20              /* Read command                    */
#define CMDwrite         0x30              /* Write command                   */
#define CMDinit          0x91              /* Init command                    */
#define CMDid            0xEC              /* Retrieve ID command             */
#define CMDdown          0xE0              /* Spin-up command                 */
#define CMDup            0xE1              /* Spin-down command               */

#define drqMask          0x04              /* Data request ready mask         */
#define busyMask         0x80              /* Status busy mask                */
#define readyMask        0x40              /* Status ready mask               */

/*                
;-------------------------------------------------------------------------------
;IDE Status Register:
;-------------------------------------------------------------------------------
;  bit 7: Busy	1=busy, 0=not busy
;  bit 6: Ready 1=ready for command, 0=not ready yet
;  bit 5: DF	1=fault occurred insIDE drive
;  bit 4: DSC	1=seek complete
;  bit 3: DRQ	1=data request ready, 0=not ready to xfer yet
;  bit 2: CORR	1=correctable error occurred
;  bit 1: IDX	vendor specific
;  bit 0: ERR	1=error occurred
;-------------------------------------------------------------------------------
*/


int debug          = FALSE;

int dPulseRst      =  50;                  /* Reset pulse length              */
int dAfterRst      =   0;                  /* Delay after reset               */
int dAfterInit     =   0;                  /* Delay after initialization      */
int dAfterSelect   =   0;                  /* Delay after drive select        */
int dBeforStatus   =   0;                  /* Delay before status             */
int dStatCommand   =   0;                  /* Status read command load        */
int dStatReadPulse =   0;                  /* Status read pulse length        */
int dStatAfterRead =   0;                  /* Delay after status read         */
int dIdCommand     =   0;                  /* Delay after ID command load     */
int dIdReadCmd     =   0;                  /* Delay after word read command   */
int dIdReadPulse   =   0;                  /* ID word read pulse length       */
int dAfterId       =   0;                  /* Delay after ID transferred      */
int dReadCommand   =   0;                  /* Read command load               */
int dReadPulse     =   0;                  /* Read pulse length               */
int dAfterRead     =   0;                  /* Delay after read                */
int dReadDeassert  =   0;                  /* Delay before pulse off          */
int dWriteCommand  =   0;                  /* Write command load              */
int dWritePulse    =   0;                  /* Write pulse length              */
int dAfterWrite    =   0;                  /* Delay after write               */


struct diskId {                                  /* IDE identification struct */
   char gCfgBinf[2];   /* General configuration bit-significant info          */
   char defCyl[2];     /* Default number of cylinders                         */
   char res1[2];       /* Reserved                                            */
   char defHd[2];      /* Default number of heads                             */
   char unfBytTr[2];   /* Number of unformatted bytes per track               */
   char unfBytSc[2];   /* Unformatted bytes per sector                        */
   char defSec[2];     /* Default number of sectors per track                 */
   char defTotSc[4];   /* Number of sectors on the device                     */
   char res2[2];       /* Reserved                                            */
   char serial[20];    /* Serial number of device (ASCII)                     */
   char bufrType[2];   /* Buffer Type (dual-ported)                           */
   char bufrSize[2];   /* Buffer size in 512-byte blocks                      */
   char eccBytes[2];   /* Number of ECC bytes passed on R/W Long commands     */
   char version[8];    /* Firmware revision (ASCII)                           */
   char model[40];     /* Device model name (ASCII)                           */
   char maxOneSc[2];   /* Maximum of one sector on R/W multiple command       */
   char dblNotSp[2];   /* Double word not supported                           */
   char advConf[2];    /* Advanced config bits: DMA bit 8, LBA bit 9          */
   char res3[2];       /* Reserved                                            */
   char pioTimng[2];   /* PIO data transfer cycle timing mode                 */
   char dmaTimng[2];   /* DMA data transfer cycle timing mode                 */
   char fieldVal[2];   /* Field validity                                      */
   char cyl[2];        /* Current number of cylinders                         */
   char hd[2];         /* Current number of heads                             */
   char sec[2];        /* Current number of sectors per track                 */
   char TotSc[4];      /* Current capacity in sectors                         */
   char MultiSec[2];   /* Multiple sector setting is valid                    */
   char TotScLBA[4];   /* Total number of sectors addressable in LBA mode     */
   char res4[4];       /* Reserved                                            */
   char advPIOmd[2];   /* Advanced PIO modes supported                        */
   char res5[4];       /* Reserved                                            */
   char minPIOnh[2];   /* Minimum PIO transfer without handshaking            */
   char minPIOrd[2];   /* Minimum PIO transfer with IORDY flow control        */
   char res6[130];     /* Reserved                                            */
   char vendor[64];    /* Reserved vendor-specific unique bytes               */
   char res7[192];     /* Reserved                                            */
};


int strncasecmp( s1, s2, length )                        /* String comparison */
   char* s1;
   char* s2;
   int length;
{
   char* c1 = s1;
   char* c2 = s2;
   int   l = 0;
   int   r = 0;

   while( (r == 0) && (l < length) && (*c2 != '\0') ) {
      r = *c1 - *c2;
      c1++;
      c2++;
      l++;
   }

   return( r );
}


int twoBytesToInt( twoBytes )        /* Convert two-bytes from IDE to integer */
   char* twoBytes;
{
   int number;
   number = (twoBytes[1] * 256) + twoBytes[0];

   return( number );
}


char* swapBytes( charArray )     /* Convert char array by swapping byte-pairs */
   char* charArray;
{
   char swapped[513];
   int  i = 0;
   int  length;

   length = strlen( charArray );

   while( i < length ) {
      swapped[i] = charArray[i+1];
      swapped[i+1] = charArray[i];
      i++;
      i++;
   }
   swapped[i] = '\0';

   return( swapped );
}


void copyBytes( ar2, ar1, n )     /* Copy "n" count of bytes from ar1 to ar2  */
   char* ar1;
   char* ar2;
   int n;
{
   int i;

   for( i=0; i < n; i++ ) {
      ar2[i] = ar1[i];
   }
   ar2[i] = '\0';
}


int hex2decimal( hex )            /* Convert a hex number to a decimal number */
   char* hex;
{
   int decimal = 0;
   int length = 0;
   int value = 0;
   int base = 1;
   int i = 0;

   length = strlen( hex );

   for( i = length--; i >= 0; i-- )
   {
      if( hex[i] >= '0' && hex[i] <= '9' )
      {
         decimal += (hex[i] - 48) * base;
         base *= 16;
      }
      else if( hex[i] >= 'A' && hex[i] <= 'F' )
      {
         decimal += (hex[i] - 55) * base;
         base *= 16;
      }
      else if( hex[i] >= 'a' && hex[i] <= 'f' )
      {
         decimal += (hex[i] - 87) * base;
         base *= 16;
      }
   }

   return( decimal );
}


int interpretNumber( number )         /* Determine if value is decimal or hex */
   char* number;
{
   int lcp;
   int value;
   char n[5];
   char lastChar;

   sprintf( n, "%s", number );

   lcp = strlen( n ) - 1;  
   lastChar = n[lcp];

   if( lastChar == 'H' || lastChar == 'h' ) {
      n[lcp] = '\0';
      value = hex2decimal( n );
   } else {
      value = (unsigned char) atoi( n );
   }

   return( value );
}


char* cleanString( string )                      /* Replace  space with  null */
   char* string;                                 /* to isolate first argument */
{
   int p;
   int l;

   l = strlen( string );
   for( p = 0; p < l; p++ ) {
      if( string[p] == ' ' ) {
        string[p] = '\0';
      }
   }

   return( string );
}


void outPort( portdata, messages )                   /* Send data to I/O port */
   char* portdata;
   int messages;
{
   unsigned char port;
   unsigned char data;
   char* op;
   char* dp;

   port   = '\0';
   data   = '\0';

   if( portdata ) {
      op = portdata;
      if ( op ) {
         if( op[0] == ' ' ) {
            op++;                                     /* remove leading space */
         }
         dp = index ( op, ' ' );
         if ( dp ) {
            if( dp[0] == ' ' ) {
               dp++;                                  /* remove leading space */
            }
            op = cleanString( op );
            port = interpretNumber( op );
            data = interpretNumber( dp );
            if( debug && messages )
               printf ( "\nPort %d (%02xH) output -> %d (%02xH)\n\n", port, port, data, data );
            out ( port, data );
         } else {
            printf ( "\nInvalid out command\n\n" ); 
         }
      } else {
         printf ( "\nInvalid out command\n\n" ); 
      }
   }
}


int IDErd8d( C )                             /* Read data from IDE interface */
   int C;
{
   char outData[20];
   int pulse;
   int delay;
   int data;

   if( debug )
      printf( "\n=== IDErd8d( %02xH ) ===\n\n", C );

   if( debug )
      printf( "Selecting IDE register %02xH.\n", C );

   sprintf( outData, "%02xH %02xH", IDEpC, C );  
   outPort( outData );

   pulse = C + IDErdLine;

   if( debug )
      printf( "Sending read pulse.\n" );
   else {
      delay = dReadCommand;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, pulse );  
   outPort( outData, verbose );

   delay = dReadPulse;
   while( delay )
      delay--;

   data = in( IDEpA );

   if( debug )
      printf( "Data returned --> %02xH <--\n", data );
   else {
      delay = dAfterRead;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, C );  
   outPort( outData, verbose );

   if( debug )
      printf( "Pulse off.\n" );
   else {
      delay = dReadDeassert;
      while( delay )
         delay--;
   }
   
   if( debug )
      printf( "Clearing 8255 port C.\n" );

   sprintf( outData, "%02xH %02xH", IDEpC, 0 );  
   outPort( outData, verbose );

   if( debug )
      printf( "==========================\n\n" );

   return( data );
}


void IDEwr8d( A, C )                            /* Send data to IDE interface */
   int A;
   int C;
{
   char outData[20];
   int pulse;
   int delay;

   if( debug )
      printf( "\n=== IDEwr8d( %02xH %02xH ) ===\n\n", A, C );

   if( debug )
      printf( "Setting 8255 to write mode.\n" );

   sprintf( outData, "%02xH %02xH", IDEpCtrl, WRITEcfg8255 );  
   outPort( outData, verbose );

   if( debug )
      printf( "Putting %02xH data in 8255 port A.\n", A );

   sprintf( outData, "%02xH %02xH", IDEpA, A );  
   outPort( outData, verbose );
   
   if( debug )
      printf( "Selecting IDE register %02xH.\n", C );

   sprintf( outData, "%02xH %02xH", IDEpC, C );  
   outPort( outData, verbose );
      
   pulse = C + IDEwrLine;

   if( debug )
      printf( "Sending write pulse.\n" );
   else {
      delay = dWriteCommand;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, pulse );  
   outPort( outData, verbose );

   if( debug )
      printf( "Pulse off.\n" );
   else {
      delay = dWritePulse;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, C );  
   outPort( outData, verbose );

   if( debug )
      printf( "Clearing 8255 port C.\n" );
   else {
      delay = dAfterWrite;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, 0 );  
   outPort( outData, verbose );

   if( debug )
      printf( "Setting 8255 control port to %02xH read mode.\n", READcfg8255 );

   sprintf( outData, "%02xH %02xH", IDEpCtrl, READcfg8255 );  
   outPort( outData, verbose );

   if( debug )
      printf( "==========================\n\n" );
}


void IDEinit() {                                  /* Initialize IDE interface */
   char outData[20];
   int delay;

   if( debug )
      printf( "\nInitializing 8255.\n" );

   sprintf( outData, "%02xH %02xH", IDEpCtrl, READcfg8255 );  
   outPort( outData, verbose );

   if( debug )
      printf( "Resetting IDE interface.\n" );

   sprintf( outData, "%02xH %02xH", IDEpC, IDErstLine );  
   outPort( outData, verbose );

   delay = dPulseRst;
   while( delay )
      delay--;

   if( debug )
      printf( "Sending reset pulse.\n" );

   sprintf( outData, "%02xH %02xH", IDEpC, 0 );  
   outPort( outData, verbose );

   if( debug )
      printf( "Putting drive into mode: Master, LBA, 512, head 0\n" );
   else {
     delay = dAfterRst;
      while( delay )
         delay--;
   }

   IDEwr8d( 0xE0, REGshd );

   delay = dAfterInit;
   while( delay )
      delay--;

}


void IDEselect( drive )                                  /* Select IDE drive */
   int drive;
{
   int delay = 0;
   char outData[20];
 
   if( debug )
      printf( "Selecting IDE unit %d.\n", drive );

   sprintf( outData, "%02xH %d", IDEDrSel, drive );  
   outPort( outData, verbose );

   delay = dAfterSelect;
   while( delay )
      delay--;
}


int getIDEstatus() {                                /* Retrieve drive status */
   int data = 0;
   int pulse = 0;
   int delay = 0;
   int rdyMask = 0x80;
   char outData[20];
 
   if( debug )
      printf( "Writing 'status request' value %02xH into 8255 port C.\n", REGstat );
   else {
      delay = dBeforStatus;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, REGstat );
   outPort( outData, verbose );

   pulse = IDErdLine + REGstat;

   if( debug )
      printf( "Now pulse the read pin by ORing %02xH with %02xH to create %02xH.\n", IDErdLine, REGstat, pulse );
   else {
      delay = dStatCommand;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, pulse );
   outPort( outData, verbose );

   if( debug )
      printf( "Reading status from IDE port A at %02xH.\n", IDEpA );
   else {
      delay = dStatReadPulse;
      while( delay )
         delay--;
   }

   data = in( IDEpA );

   if( debug )
      printf( "\nStatus --> %02xH <--\n\n", data );

   if( debug )
      printf( "De-assert read pin, resetting Port C to value %02xH.\n", REGstat );

   sprintf( outData, "%02xH %02xH", IDEpC, REGstat );
   outPort( outData, verbose );

   if( debug )
      printf( "Set Port C to value %02xH.\n", 0 );
   else {
      delay = dStatAfterRead;
      while( delay )
         delay--;
   }

   sprintf( outData, "%02xH %02xH", IDEpC, 0 );
   outPort( outData, verbose );

   return( data );
}


void getIDEid( drive, status )       /* Retrieve drive identification string */
   int drive;
   int status;
{
   int i;
   int j;
   int data = 0x00;
   int pulse = 0x00;
   int delay = 0x00;
   unsigned char lb;
   unsigned char hb;
   char outData[20];
   char ident[513];
   char drdata[513];
   char model[21];
   char serial[11];
   char version[5];
   int  cylinders;
   int  heads;
   int  sectors;

   struct diskId* diskInfo = (struct diskId*) ident;

   if( status & readyMask ) {

      if( debug )
         printf( "Requesting drive information.\n" );

      IDEwr8d( CMDid, REGcmd );

      delay = dIdCommand;
      while( delay )
         delay--;

      data = IDErd8d( REGstat );

      if( debug ) {
         printf( "Drive %d returned status code %02xH, after requesting identification.\n", drive, data );     
         printf( "(Note: 58H means data is ready to be retrieved from the drive.)\n" );
      }

      if( data == 0x58 ) {

         if( debug )
            printf( "\n===============================================================================\n" );

         for( i = 0; i < 256; i++ ) {
            j = i * 2;
            sprintf( outData, "%02xH %02xH", IDEpC, REGdata );
            outPort( outData, quiet );

            delay = dIdReadCmd;
               while( delay )
                  delay--;

            pulse = REGdata + IDErdLine;

            sprintf( outData, "%02xH %02xH", IDEpC, pulse );
            outPort( outData, quiet );

            delay = dIdReadPulse;
               while( delay )
                  delay--;

            ident[j] = in( IDEpA );                             /* lower byte */
            ident[j+1] = in( IDEpB );                           /* upper byte */

            if( debug ) {
               printf( "%c", ident[j] );
               printf( "%c", ident[j+1] );
            }

            sprintf( outData, "%02xH %02xH", IDEpC, REGdata );
            outPort( outData, quiet );

            delay = dAfterId;
               while( delay )
                  delay--;
         }
         j++;
         j++;
         ident[j] = '\0';

         if( debug )
            printf( "\n===============================================================================\n" );

         printf( "\nDrive %d identifies itself as:\n\n", drive );

         copyBytes( drdata, diskInfo->model, 40 );
         sprintf( model, "%s", swapBytes(drdata) );
         printf( "Model:     %s\n", model );

         copyBytes( drdata, diskInfo->serial, 20 );
         sprintf( serial, "%s", drdata );
         printf( "Serial:    %s\n", serial );

         copyBytes( drdata, diskInfo->version, 8 );
         sprintf( version, "%s", swapBytes(drdata) );
         printf( "Version:   %s\n", version );

         copyBytes( drdata, diskInfo->cyl, 2 );
         cylinders = twoBytesToInt( drdata );
         printf( "Cylinders: %d\n", cylinders );
  
         copyBytes( drdata, diskInfo->hd, 2 );
         heads = twoBytesToInt( drdata );
         printf( "Heads:     %d\n", heads );

         copyBytes( drdata, diskInfo->sec, 2 );
         sectors = twoBytesToInt( drdata );
         printf( "Sectors:   %d\n", sectors );

      } else {

         printf( "Not attempting drive identification because of unexpected status.\n" );     
      }

      if( debug ) {
         status = getIDEstatus();
         printf( "\nCompleted with status code %02xH.\n", status );
      }     

   } else {

      printf( "Drive %d busy with status code %02xH.\n", drive, status );
   }

}


void clearScreen() {                                       /* Scroll 25 lines */
   int l;
   putchar ( '\r' );
   for ( l = 0; l < 25; l++ ) {
      putchar ( '\n' );
   }
}


int main() {
   unsigned char port;
   unsigned char data;
   unsigned char val;
   int showdata;
   int offset;
   int going;
   int instr;
   int drive;
   int start;
   int stat;
   int end;
   int top;
   int cls;
   int l;
   int v;
   char* pd;
   char cmd[80];

   val    = '\0';
   cmd[0] = '\0';
   showdata = 0;
   going  = TRUE;
   instr  = FALSE;
   start  = FALSE;
   cls    = TRUE;
   end    = 16;
   top    = 127;
   offset = 0;
   l      = 0;
   v      = 1;


   while ( going ) {

      if ( cls ) {
         clearScreen();
      }

      if ( showdata ) {
         cls = TRUE;
         puts ("================================= Input Data ==================================");

         while ( start < top ) {

            for ( l = start; l < end; l++ ) {
               data = in( l );
               printf( "P0%02x", l );
               if ( l < (end - 1) ) {
                  printf( " " );
               }
            }
            printf( "\n\r" );

            printf( " " );
            for ( l = start; l < end; l++ ) {
               data = in( l );
               printf( "%02x", data );
               if ( l < (end - 1) ) {
                  printf( "   " );
               }
            }
            printf( "\n\r" );

            start += 16;
            end += 16;
         }
         puts ("===============================================================================");
      }

      if ( instr ) {
         cls = TRUE;
         puts ("===============================================================================");
         puts ("The  Intel 8080 processor provides 256 addresses for input and output  devices.");
         puts ("This  program  exposes all I/O addresses, allowing you to see the  data  placed");
         puts ("onto  each  of the ports.   You can also send a byte out any of  the  I/O  port");
         puts ("addresses.  Of course, if there is no hardware addressed on a port,  then  data");
         puts ("read will be undetermined.  Data written to an unused port will have no effect.");
         puts ("                                                                               ");
         puts ("Available  commands  are help, show, view, out, debug, test,  idetest,  testide");
         puts ("and quit.  Help shows this message.   Quit ends the program.   Show and view do");
         puts ("the  same thing, showing the top or bottom 128 input ports.   One  argument  is");
         puts ("expected  on the show or view command - either \"high\" or \"low\" - which  can  be");
         puts ("abbreviated.");
         puts ("                                                                               ");
         puts ("Test,  idetest and testide all do the same thing, which is to check for an  IDE");
         puts ("subsystem and to report the drive status information, if present.  An  optional");
         puts ("argument - 0 or 1 - can be entered to select the drive number.");
         puts ("                                                                               ");
         puts ("The  out command requires two arguments, first the port number and second,  the");
         puts ("data  that will be presented to the port.   Both arguments are expected  to  be");
         puts ("one  byte  wide,  e.g. 0 to 255.   Numeric arguments can be either  decimal  or");
         puts ("hexadecimal.   Hex values are indicated by a trailing \"H\" character, which must");
         puts ("be present even if some digits are alpha characters.                           ");
         puts ("===============================================================================");
      }

      printf ("\nCommand:  ");

      fgets ( cmd, sizeof(cmd)-1, stdin );
      cmd[strlen(cmd)-1] = '\0';

      if ( strncasecmp(cmd, "quit", 4) == 0 ) {
         going = FALSE;  
      }
      else if ( strncasecmp(cmd, "help", 4) == 0 ) {
         showdata = FALSE;
         instr = TRUE;
      }
      else if ( strncasecmp(cmd, "show h", 6) == 0 ) {
         showdata = TRUE;
         instr = FALSE;
         v = 0;         
      }
      else if ( strncasecmp(cmd, "show l", 6) == 0 ) {
         showdata = TRUE;
         instr = FALSE;
         v = 1;         
      }
      else if ( strncasecmp(cmd, "view h", 6) == 0 ) {
         showdata = TRUE;
         v = 0;         
      }
      else if ( strncasecmp(cmd, "view l", 6) == 0 ) {
         showdata = TRUE;
         v = 1;         
      }
      else if ( strncasecmp(cmd, "out", 3) == 0 ) {
         showdata = 0;
         instr = 0;
         cls = 0;
         pd = index ( cmd, ' ' );
         outPort( pd );
      }
      else if ( strncasecmp(cmd, "testide", 7) == 0 ) {
         showdata = FALSE;
         instr = FALSE;
         cls = FALSE;
         pd = index ( cmd, ' ' );
         drive = atoi( pd );
         if( (drive == 0) || (drive == 1) ) {
            printf( "Examining drive %d...\n", drive );
            IDEinit();
            IDEselect( drive );
            stat = getIDEstatus();
            getIDEid( drive, stat );
         } else {
            printf( "Unsupported drive %d.\n", drive );
         }        
      }
      else if ( strncasecmp(cmd, "idetest", 7) == 0 ) {
         showdata = FALSE;
         instr = FALSE;
         cls = FALSE;
         pd = index ( cmd, ' ' );
         drive = atoi( pd );
         if( (drive == 0) || (drive == 1) ) {
            printf( "Examining drive %d...\n", drive );
            IDEinit();
            IDEselect( drive );
            stat = getIDEstatus();
            getIDEid( drive, stat );
         } else {
            printf( "Unsupported drive %d.\n", drive );
         }        
      }
      else if ( strncasecmp(cmd, "test", 4) == 0 ) {
         showdata = FALSE;
         instr = FALSE;
         cls = FALSE;
         pd = index ( cmd, ' ' );
         drive = atoi( pd );
         if( (drive == 0) || (drive == 1) ) {
            printf( "Examining drive %d...\n", drive );
            IDEinit();
            IDEselect( drive );
            stat = getIDEstatus();
            getIDEid( drive, stat );
         } else {
            printf( "Unsupported drive %d.\n", drive );
         }        
      }
      else if ( strncasecmp(cmd, "debug on", 8) == 0 ) {
         debug = TRUE;
         printf( "Debug is turned ON.\n" );
      }
      else if ( strncasecmp(cmd, "debug off", 8) == 0 ) {
         debug = FALSE;
         printf( "Debug is turned OFF.\n" );
      }
      else if ( strncasecmp(cmd, "dpulserst", 9) == 0 ) {
         pd = index ( cmd, ' ' );
         dPulseRst = atoi( pd );
         printf( "dPulseRst value is now set to %d.\n", dPulseRst );
      }
      else if ( strncasecmp(cmd, "dafterrst", 9) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterRst = atoi( pd );
         printf( "dAfterRst value is now set to %d.\n", dAfterRst );
      }
      else if ( strncasecmp(cmd, "dpulsereset", 11) == 0 ) {
         pd = index ( cmd, ' ' );
         dPulseRst = atoi( pd );
         printf( "dPulseRst value is now set to %d.\n", dPulseRst );
      }
      else if ( strncasecmp(cmd, "dafterreset", 11) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterRst = atoi( pd );
         printf( "dAfterRst value is now set to %d.\n", dAfterRst );
      }
      else if ( strncasecmp(cmd, "dafterinit", 10) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterInit = atoi( pd );
         printf( "dAfterInit value is now set to %d.\n", dAfterInit );
      }
      else if ( strncasecmp(cmd, "dafterselect", 12) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterSelect = atoi( pd );
         printf( "dAfterSelect value is now set to %d.\n", dAfterSelect );
      }
      else if ( strncasecmp(cmd, "dbeforstatus", 12) == 0 ) {
         pd = index ( cmd, ' ' );
         dBeforStatus = atoi( pd );
         printf( "dBeforStatus value is now set to %d.\n", dBeforStatus );
      }
      else if ( strncasecmp(cmd, "dbeforestatus", 13) == 0 ) {
         pd = index ( cmd, ' ' );
         dBeforStatus = atoi( pd );
         printf( "dBeforStatus value is now set to %d.\n", dBeforStatus );
      }
      else if ( strncasecmp(cmd, "dstatcommand", 12) == 0 ) {
         pd = index ( cmd, ' ' );
         dStatCommand = atoi( pd );
         printf( "dStatCommand value is now set to %d.\n", dStatCommand );
      }
      else if ( strncasecmp(cmd, "dstatreadpulse", 14) == 0 ) {
         pd = index ( cmd, ' ' );
         dStatReadPulse = atoi( pd );
         printf( "dStatReadPulse value is now set to %d.\n", dStatReadPulse );
      }
      else if ( strncasecmp(cmd, "dstatafterread", 14) == 0 ) {
         pd = index ( cmd, ' ' );
         dStatAfterRead = atoi( pd );
         printf( "dStatAfterRead value is now set to %d.\n", dStatAfterRead );
      }
      else if ( strncasecmp(cmd, "didcommand", 10) == 0 ) {
         pd = index ( cmd, ' ' );
         dIdCommand = atoi( pd );
         printf( "dIdCommand value is now set to %d.\n", dIdCommand );
      }
      else if ( strncasecmp(cmd, "didreadcmd", 10) == 0 ) {
         pd = index ( cmd, ' ' );
         dIdReadCmd = atoi( pd );
         printf( "dIdReadCmd value is now set to %d.\n", dIdReadCmd );
      }
      else if ( strncasecmp(cmd, "didreadpulse", 12) == 0 ) {
         pd = index ( cmd, ' ' );
         dIdReadPulse = atoi( pd );
         printf( "dIdReadPulse value is now set to %d.\n", dIdReadPulse );
      }
      else if ( strncasecmp(cmd, "dafterid", 8) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterId = atoi( pd );
         printf( "dAfterId value is now set to %d.\n", dAfterId );
      }
      else if ( strncasecmp(cmd, "dreadcommand", 12) == 0 ) {
         pd = index ( cmd, ' ' );
         dReadCommand = atoi( pd );
         printf( "dReadCommand value is now set to %d.\n", dReadCommand );
      }
      else if ( strncasecmp(cmd, "dreadpulse", 10) == 0 ) {
         pd = index ( cmd, ' ' );
         dReadPulse = atoi( pd );
         printf( "dReadPulse value is now set to %d.\n", dReadPulse );
      }
      else if ( strncasecmp(cmd, "dafterread", 10) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterRead = atoi( pd );
         printf( "dAfterRead value is now set to %d.\n", dAfterRead );
      }
      else if ( strncasecmp(cmd, "dreaddeassert", 13) == 0 ) {
         pd = index ( cmd, ' ' );
         dReadDeassert = atoi( pd );
         printf( "dReadDeassert value is now set to %d.\n", dReadDeassert );
      }
      else if ( strncasecmp(cmd, "dwritecommand", 13) == 0 ) {
         pd = index ( cmd, ' ' );
         dWriteCommand = atoi( pd );
         printf( "dWriteCommand value is now set to %d.\n", dWriteCommand );
      }
      else if ( strncasecmp(cmd, "dwritepulse", 11) == 0 ) {
         pd = index ( cmd, ' ' );
         dWritePulse = atoi( pd );
         printf( "dWritePulse value is now set to %d.\n", dWritePulse );
      }
      else if ( strncasecmp(cmd, "dafterwrite", 11) == 0 ) {
         pd = index ( cmd, ' ' );
         dAfterWrite = atoi( pd );
         printf( "dAfterWrite value is now set to %d.\n", dAfterWrite );
      }

      if ( v ) {
         start = 0;  /* view bottom 128 input ports */
         end = 16;
         top = 127;  
      } else {       /* view top 128 input ports */
         start = 128;
         end = 144;
         top = 255;  
      }
   }

   puts ("\n\nEnjoy your day!\n\n");

   return( 0 );
}
